@@ -11,6 +11,8 @@
 #ifdef PGRN_SUPPORT_TABLEAM
 #	include <access/tableam.h>
 #endif
+#include <catalog/pg_type.h>
+#include <funcapi.h>
 #include <miscadmin.h>
 
 static bool PGrnWALEnabled = false;
@@ -63,6 +65,7 @@ PGrnWALSetMaxSize(size_t size)
 
 PGDLLEXPORT PG_FUNCTION_INFO_V1(pgroonga_wal_apply_index);
 PGDLLEXPORT PG_FUNCTION_INFO_V1(pgroonga_wal_apply_all);
+PGDLLEXPORT PG_FUNCTION_INFO_V1(pgroonga_wal_status);
 PGDLLEXPORT PG_FUNCTION_INFO_V1(pgroonga_wal_truncate_index);
 PGDLLEXPORT PG_FUNCTION_INFO_V1(pgroonga_wal_truncate_all);
 
@@ -209,15 +212,6 @@ msgpack_pack_grn_obj(msgpack_packer *packer, grn_obj *object)
 #define PGRN_WAL_META_PAGE_BLOCK_NUMBER 0
 
 #ifdef PGRN_SUPPORT_WAL
-static LOCKMODE
-PGrnWALLockMode(void)
-{
-	if (RecoveryInProgress())
-		return RowExclusiveLock;
-	else
-		return ShareUpdateExclusiveLock;
-}
-
 static Buffer
 PGrnWALReadLockedBuffer(Relation index,
 						BlockNumber blockNumber,
@@ -236,6 +230,30 @@ PGrnWALReadLockedBuffer(Relation index,
 	return buffer;
 }
 
+static char *
+PGrnWALPageGetData(Page page)
+{
+	return PageGetContents(page);
+}
+
+static size_t
+PGrnWALPageGetFreeSize(Page page)
+{
+	PageHeader pageHeader;
+
+	pageHeader = (PageHeader)page;
+	return pageHeader->pd_upper - pageHeader->pd_lower;
+}
+
+static LocationIndex
+PGrnWALPageGetLastOffset(Page page)
+{
+	PageHeader pageHeader;
+
+	pageHeader = (PageHeader)page;
+	return pageHeader->pd_lower - SizeOfPageHeaderData;
+}
+
 static void
 PGrnWALDataInitBuffers(PGrnWALData *data)
 {
@@ -307,10 +325,29 @@ PGrnWALDataInitCurrent(PGrnWALData *data)
 	data->current.page = NULL;
 }
 
+static void
+PGrnWALDataFinish(PGrnWALData *data)
+{
+	BlockNumber block;
+	LocationIndex offset;
+	if (data->current.page)
+	{
+		block = BufferGetBlockNumber(data->current.buffer);
+		offset = PGrnWALPageGetLastOffset(data->current.page);
+	}
+	else
+	{
+		block = data->meta.pageSpecial->next;
+		offset = 0;
+	}
+	GenericXLogFinish(data->state);
+	PGrnIndexStatusSetWALAppliedPosition(data->index, block, offset);
+}
+
 static void
 PGrnWALDataRestart(PGrnWALData *data)
 {
-	GenericXLogFinish(data->state);
+	PGrnWALDataFinish(data);
 
 	PGrnWALDataReleaseBuffers(data);
 
@@ -320,30 +357,6 @@ PGrnWALDataRestart(PGrnWALData *data)
 	PGrnWALDataInitCurrent(data);
 }
 
-static char *
-PGrnWALPageGetData(Page page)
-{
-	return PageGetContents(page);
-}
-
-static size_t
-PGrnWALPageGetFreeSize(Page page)
-{
-	PageHeader pageHeader;
-
-	pageHeader = (PageHeader)page;
-	return pageHeader->pd_upper - pageHeader->pd_lower;
-}
-
-static LocationIndex
-PGrnWALPageGetLastOffset(Page page)
-{
-	PageHeader pageHeader;
-
-	pageHeader = (PageHeader)page;
-	return pageHeader->pd_lower - SizeOfPageHeaderData;
-}
-
 static void
 PGrnWALPageAppend(Page page, const char *data, size_t dataSize)
 {
@@ -454,10 +467,6 @@ PGrnWALPageWriter(void *userData,
 		if (rest <= freeSize)
 		{
 			PGrnWALPageAppend(data->current.page, buffer, rest);
-			PGrnIndexStatusSetWALAppliedPosition(
-				data->index,
-				BufferGetBlockNumber(data->current.buffer),
-				PGrnWALPageGetLastOffset(data->current.page));
 			written += rest;
 		}
 		else
@@ -480,6 +489,36 @@ PGrnWALDataInitMessagePack(PGrnWALData *data)
 {
 	msgpack_packer_init(&(data->packer), data, PGrnWALPageWriter);
 }
+
+static LOCKMODE
+PGrnWALLockMode(void)
+{
+	if (RecoveryInProgress())
+		return RowExclusiveLock;
+	else
+		return ShareUpdateExclusiveLock;
+}
+
+static BlockNumber
+PGrnWALLockBlockNumber(void)
+{
+	/* We can use any block number for this. We just want an index
+	 * level lock but we can't use LockRelation(index) because it
+	 * conflicts with REINDEX INDEX CONCURRENTLY. */
+	return 0;
+}
+
+static void
+PGrnWALLock(Relation index)
+{
+	LockPage(index, PGrnWALLockBlockNumber(), PGrnWALLockMode());
+}
+
+static void
+PGrnWALUnlock(Relation index)
+{
+	UnlockPage(index, PGrnWALLockBlockNumber(), PGrnWALLockMode());
+}
 #endif
 
 PGrnWALData *
@@ -494,7 +533,7 @@ PGrnWALStart(Relation index)
 	if (!RelationIsValid(index))
 		return NULL;
 
-	LockRelation(index, PGrnWALLockMode());
+	PGrnWALLock(index);
 
 	data = palloc(sizeof(PGrnWALData));
 
@@ -520,11 +559,11 @@ PGrnWALFinish(PGrnWALData *data)
 	if (!data)
 		return;
 
-	GenericXLogFinish(data->state);
+	PGrnWALDataFinish(data);
 
 	PGrnWALDataReleaseBuffers(data);
 
-	UnlockRelation(data->index, PGrnWALLockMode());
+	PGrnWALUnlock(data->index);
 
 	pfree(data);
 #endif
@@ -548,7 +587,7 @@ PGrnWALAbort(PGrnWALData *data)
 	{
 		PGrnWALDataReleaseBuffers(data);
 
-		UnlockRelation(data->index, PGrnWALLockMode());
+		PGrnWALUnlock(data->index);
 	}
 
 	pfree(data);
@@ -1120,7 +1159,8 @@ PGrnWALApplyNeeded(PGrnWALApplyData *data)
 }
 
 static bool
-PGrnWALApplyKeyEqual(const char *context,
+PGrnWALApplyKeyEqual(PGrnWALApplyData *data,
+					 const char *context,
 					 msgpack_object *key,
 					 const char *name)
 {
@@ -1130,9 +1170,11 @@ PGrnWALApplyKeyEqual(const char *context,
 	if (key->type != MSGPACK_OBJECT_STR)
 	{
 		PGrnCheckRC(GRN_INVALID_ARGUMENT,
-					"%s%s%skey must be string: <%#x>",
+					"%s%s[%s(%u)]%skey must be string: <%#x>",
 					tag,
 					context ? context : "",
+					RelationGetRelationName(data->index),
+					RelationGetRelid(data->index),
 					context ? " " : "",
 					key->type);
 	}
@@ -1147,16 +1189,19 @@ PGrnWALApplyKeyEqual(const char *context,
 }
 
 static uint64_t
-PGrnWALApplyValueGetPositiveInteger(const char *context,
+PGrnWALApplyValueGetPositiveInteger(PGrnWALApplyData *data,
+									const char *context,
 									msgpack_object_kv *kv)
 {
 	const char *tag = "[wal][apply][value][positive-integer][get]";
 	if (kv->val.type != MSGPACK_OBJECT_POSITIVE_INTEGER)
 	{
 		PGrnCheckRC(GRN_INVALID_ARGUMENT,
-					"%s%s%s%.*s value must be positive integer: <%#x>",
+					"%s%s[%s(%u)]%s%.*s value must be positive integer: <%#x>",
 					tag,
 					context ? context : "",
+					RelationGetRelationName(data->index),
+					RelationGetRelid(data->index),
 					context ? " " : "",
 					MSGPACK_OBJECT_VIA_STR(kv->key).size,
 					MSGPACK_OBJECT_VIA_STR(kv->key).ptr,
@@ -1167,7 +1212,8 @@ PGrnWALApplyValueGetPositiveInteger(const char *context,
 }
 
 static void
-PGrnWALApplyValueGetString(const char *context,
+PGrnWALApplyValueGetString(PGrnWALApplyData *data,
+						   const char *context,
 						   msgpack_object_kv *kv,
 						   const char **string,
 						   size_t *stringSize)
@@ -1176,9 +1222,11 @@ PGrnWALApplyValueGetString(const char *context,
 	if (kv->val.type != MSGPACK_OBJECT_STR)
 	{
 		PGrnCheckRC(GRN_INVALID_ARGUMENT,
-					"%s%s%s%.*s value must be string: <%#x>",
+					"%s%s[%s(%u)]%s%.*s value must be string: <%#x>",
 					tag,
 					context ? context : "",
+					RelationGetRelationName(data->index),
+					RelationGetRelid(data->index),
 					context ? " " : "",
 					MSGPACK_OBJECT_VIA_STR(kv->key).size,
 					MSGPACK_OBJECT_VIA_STR(kv->key).ptr,
@@ -1190,7 +1238,8 @@ PGrnWALApplyValueGetString(const char *context,
 }
 
 static void
-PGrnWALApplyValueGetBinary(const char *context,
+PGrnWALApplyValueGetBinary(PGrnWALApplyData *data,
+						   const char *context,
 						   msgpack_object_kv *kv,
 						   const char **binary,
 						   size_t *binarySize)
@@ -1199,9 +1248,11 @@ PGrnWALApplyValueGetBinary(const char *context,
 	if (kv->val.type != MSGPACK_OBJECT_BIN)
 	{
 		PGrnCheckRC(GRN_INVALID_ARGUMENT,
-					"%s%s%s%.*s value must be binary: <%#x>",
+					"%s%s[%s(%u)]%s%.*s value must be binary: <%#x>",
 					tag,
 					context ? context : "",
+					RelationGetRelationName(data->index),
+					RelationGetRelid(data->index),
 					context ? " " : "",
 					MSGPACK_OBJECT_VIA_BIN(kv->key).size,
 					MSGPACK_OBJECT_VIA_BIN(kv->key).ptr,
@@ -1213,7 +1264,8 @@ PGrnWALApplyValueGetBinary(const char *context,
 }
 
 static grn_obj *
-PGrnWALApplyValueGetGroongaObject(const char *context,
+PGrnWALApplyValueGetGroongaObject(PGrnWALApplyData *data,
+								  const char *context,
 								  msgpack_object_kv *kv)
 {
 	const char *tag = "[wal][apply][value][groonga-object][get]";
@@ -1231,9 +1283,11 @@ PGrnWALApplyValueGetGroongaObject(const char *context,
 		break;
 	default:
 		PGrnCheckRC(GRN_INVALID_ARGUMENT,
-					"%s%s%s%.*s value must be nil or string: <%#x>",
+					"%s%s[%s(%u)]%s%.*s value must be nil or string: <%#x>",
 					tag,
 					context ? context : "",
+					RelationGetRelationName(data->index),
+					RelationGetRelid(data->index),
 					context ? " " : "",
 					MSGPACK_OBJECT_VIA_STR(kv->key).size,
 					MSGPACK_OBJECT_VIA_STR(kv->key).ptr,
@@ -1245,7 +1299,8 @@ PGrnWALApplyValueGetGroongaObject(const char *context,
 }
 
 static void
-PGrnWALApplyValueGetGroongaObjectIDs(const char *context,
+PGrnWALApplyValueGetGroongaObjectIDs(PGrnWALApplyData *data,
+									 const char *context,
 									 msgpack_object_kv *kv,
 									 grn_obj *ids)
 {
@@ -1256,9 +1311,11 @@ PGrnWALApplyValueGetGroongaObjectIDs(const char *context,
 	if (kv->val.type != MSGPACK_OBJECT_ARRAY)
 	{
 		PGrnCheckRC(GRN_INVALID_ARGUMENT,
-					"%s%s%s%.*s value must be array: <%#x>",
+					"%s%s[%s(%u)]%s%.*s value must be array: <%#x>",
 					tag,
 					context ? context : "",
+					RelationGetRelationName(data->index),
+					RelationGetRelid(data->index),
 					context ? " " : "",
 					MSGPACK_OBJECT_VIA_STR(kv->key).size,
 					MSGPACK_OBJECT_VIA_STR(kv->key).ptr,
@@ -1276,9 +1333,12 @@ PGrnWALApplyValueGetGroongaObjectIDs(const char *context,
 		if (element->type != MSGPACK_OBJECT_STR)
 		{
 			PGrnCheckRC(GRN_INVALID_ARGUMENT,
-						"%s%s%s%.*s value must be array of string: [%u]=<%#x>",
+						"%s%s[%s(%u)]%s%.*s "
+						"value must be array of string: [%u]=<%#x>",
 						tag,
 						context ? context : "",
+						RelationGetRelationName(data->index),
+						RelationGetRelid(data->index),
 						context ? " " : "",
 						MSGPACK_OBJECT_VIA_STR(kv->key).size,
 						MSGPACK_OBJECT_VIA_STR(kv->key).ptr,
@@ -1295,7 +1355,8 @@ PGrnWALApplyValueGetGroongaObjectIDs(const char *context,
 }
 
 static grn_obj *
-PGrnWALApplyValueGetTableModule(const char *context,
+PGrnWALApplyValueGetTableModule(PGrnWALApplyData *data,
+								const char *context,
 								msgpack_object_kv *kv,
 								grn_obj *buffer)
 {
@@ -1316,9 +1377,11 @@ PGrnWALApplyValueGetTableModule(const char *context,
 		break;
 	default:
 		PGrnCheckRC(GRN_INVALID_ARGUMENT,
-					"%s%s%s%.*s value must be nil or string: <%#x>",
+					"%s%s[%s(%u)]%s%.*s value must be nil or string: <%#x>",
 					tag,
 					context ? context : "",
+					RelationGetRelationName(data->index),
+					RelationGetRelid(data->index),
 					context ? " " : "",
 					MSGPACK_OBJECT_VIA_STR(kv->key).size,
 					MSGPACK_OBJECT_VIA_STR(kv->key).ptr,
@@ -1330,7 +1393,8 @@ PGrnWALApplyValueGetTableModule(const char *context,
 }
 
 static grn_obj *
-PGrnWALApplyValueGetTableModules(const char *context,
+PGrnWALApplyValueGetTableModules(PGrnWALApplyData *data,
+								 const char *context,
 								 msgpack_object_kv *kv,
 								 grn_obj *buffer)
 {
@@ -1364,11 +1428,13 @@ PGrnWALApplyValueGetTableModules(const char *context,
 			if (element->type != MSGPACK_OBJECT_STR)
 			{
 				PGrnCheckRC(GRN_INVALID_ARGUMENT,
-							"%s%s%s"
+							"%s%s[%s(%u)]%s"
 							"%.*s value must be string or array of string: "
 							"[%u]=<%#x>",
 							tag,
 							context ? context : "",
+							RelationGetRelationName(data->index),
+							RelationGetRelid(data->index),
 							context ? " " : "",
 							MSGPACK_OBJECT_VIA_STR(kv->key).size,
 							MSGPACK_OBJECT_VIA_STR(kv->key).ptr,
@@ -1387,9 +1453,11 @@ PGrnWALApplyValueGetTableModules(const char *context,
 	}
 	default:
 		PGrnCheckRC(GRN_INVALID_ARGUMENT,
-					"%s%s%s%.*s value must be string or array: <%#x>",
+					"%s%s[%s(%u)]%s%.*s value must be string or array: <%#x>",
 					tag,
 					context ? context : "",
+					RelationGetRelationName(data->index),
+					RelationGetRelid(data->index),
 					context ? " " : "",
 					MSGPACK_OBJECT_VIA_STR(kv->key).size,
 					MSGPACK_OBJECT_VIA_STR(kv->key).ptr,
@@ -1497,8 +1565,10 @@ PGrnWALApplyInsertArray(PGrnWALApplyData *data,
 			break;
 		default:
 			PGrnCheckRC(GRN_INVALID_ARGUMENT,
-						"%s unexpected element type: <%#x>",
+						"%s[%s(%u)] unexpected element type: <%#x>",
 						tag,
+						RelationGetRelationName(data->index),
+						RelationGetRelid(data->index),
 						element->type);
 			break;
 		}
@@ -1523,9 +1593,9 @@ PGrnWALApplyInsert(PGrnWALApplyData *data,
 		msgpack_object_kv *kv;
 
 		kv = &(map->ptr[currentElement]);
-		if (PGrnWALApplyKeyEqual(context, &(kv->key), "_table"))
+		if (PGrnWALApplyKeyEqual(data, context, &(kv->key), "_table"))
 		{
-			table = PGrnWALApplyValueGetGroongaObject(context, kv);
+			table = PGrnWALApplyValueGetGroongaObject(data, context, kv);
 			currentElement++;
 		}
 	}
@@ -1541,9 +1611,9 @@ PGrnWALApplyInsert(PGrnWALApplyData *data,
 		msgpack_object_kv *kv;
 
 		kv = &(map->ptr[currentElement]);
-		if (PGrnWALApplyKeyEqual(context, &(kv->key), GRN_COLUMN_NAME_KEY))
+		if (PGrnWALApplyKeyEqual(data, context, &(kv->key), GRN_COLUMN_NAME_KEY))
 		{
-			PGrnWALApplyValueGetBinary(context, kv, &key, &keySize);
+			PGrnWALApplyValueGetBinary(data, context, kv, &key, &keySize);
 			currentElement++;
 		}
 	}
@@ -1562,9 +1632,11 @@ PGrnWALApplyInsert(PGrnWALApplyData *data,
 		if (key->type != MSGPACK_OBJECT_STR)
 		{
 			PGrnCheckRC(GRN_INVALID_ARGUMENT,
-						"%s%s key must be map: <%#x>",
+						"%s%s[%s(%u)] key must be map: <%#x>",
 						tag,
 						context,
+						RelationGetRelationName(data->index),
+						RelationGetRelid(data->index),
 						key->type);
 		}
 
@@ -1612,9 +1684,11 @@ PGrnWALApplyInsert(PGrnWALApplyData *data,
 */
 		default:
 			PGrnCheckRC(GRN_INVALID_ARGUMENT,
-						"%s%s unexpected value type: <%#x>",
+						"%s%s[%s(%u)] unexpected value type: <%#x>",
 						tag,
 						context,
+						RelationGetRelationName(data->index),
+						RelationGetRelid(data->index),
 						value->type);
 			break;
 		}
@@ -1642,34 +1716,37 @@ PGrnWALApplyCreateTable(PGrnWALApplyData *data,
 		msgpack_object_kv *kv;
 
 		kv = &(map->ptr[i]);
-		if (PGrnWALApplyKeyEqual(context, &(kv->key), "name"))
+		if (PGrnWALApplyKeyEqual(data, context, &(kv->key), "name"))
 		{
-			PGrnWALApplyValueGetString(context, kv, &name, &nameSize);
+			PGrnWALApplyValueGetString(data, context, kv, &name, &nameSize);
 		}
-		else if (PGrnWALApplyKeyEqual(context, &(kv->key), "flags"))
+		else if (PGrnWALApplyKeyEqual(data, context, &(kv->key), "flags"))
 		{
-			flags = PGrnWALApplyValueGetPositiveInteger(context, kv);
+			flags = PGrnWALApplyValueGetPositiveInteger(data, context, kv);
 		}
-		else if (PGrnWALApplyKeyEqual(context, &(kv->key), "type"))
+		else if (PGrnWALApplyKeyEqual(data, context, &(kv->key), "type"))
 		{
-			type = PGrnWALApplyValueGetGroongaObject(context, kv);
+			type = PGrnWALApplyValueGetGroongaObject(data, context, kv);
 		}
-		else if (PGrnWALApplyKeyEqual(context, &(kv->key), "tokenizer"))
+		else if (PGrnWALApplyKeyEqual(data, context, &(kv->key), "tokenizer"))
 		{
-			tokenizer = PGrnWALApplyValueGetTableModule(context,
+			tokenizer = PGrnWALApplyValueGetTableModule(data,
+														context,
 														kv,
 														&(buffers->tokenizer));
 		}
-		else if (PGrnWALApplyKeyEqual(context, &(kv->key), "normalizer") ||
-				 PGrnWALApplyKeyEqual(context, &(kv->key), "normalizers"))
+		else if (PGrnWALApplyKeyEqual(data, context, &(kv->key), "normalizer") ||
+				 PGrnWALApplyKeyEqual(data, context, &(kv->key), "normalizers"))
 		{
-			normalizers = PGrnWALApplyValueGetTableModule(context,
+			normalizers = PGrnWALApplyValueGetTableModule(data,
+														  context,
 														  kv,
 														  &(buffers->normalizers));
 		}
-		else if (PGrnWALApplyKeyEqual(context, &(kv->key), "token_filters"))
+		else if (PGrnWALApplyKeyEqual(data, context, &(kv->key), "token_filters"))
 		{
-			tokenFilters = PGrnWALApplyValueGetTableModules(context,
+			tokenFilters = PGrnWALApplyValueGetTableModules(data,
+															context,
 															kv,
 															&(buffers->tokenFilters));
 		}
@@ -1703,21 +1780,21 @@ PGrnWALApplyCreateColumn(PGrnWALApplyData *data,
 		msgpack_object_kv *kv;
 
 		kv = &(map->ptr[i]);
-		if (PGrnWALApplyKeyEqual(context, &(kv->key), "table"))
+		if (PGrnWALApplyKeyEqual(data, context, &(kv->key), "table"))
 		{
-			table = PGrnWALApplyValueGetGroongaObject(context, kv);
+			table = PGrnWALApplyValueGetGroongaObject(data, context, kv);
 		}
-		else if (PGrnWALApplyKeyEqual(context, &(kv->key), "name"))
+		else if (PGrnWALApplyKeyEqual(data, context, &(kv->key), "name"))
 		{
-			PGrnWALApplyValueGetString(context, kv, &name, &nameSize);
+			PGrnWALApplyValueGetString(data, context, kv, &name, &nameSize);
 		}
-		else if (PGrnWALApplyKeyEqual(context, &(kv->key), "flags"))
+		else if (PGrnWALApplyKeyEqual(data, context, &(kv->key), "flags"))
 		{
-			flags = PGrnWALApplyValueGetPositiveInteger(context, kv);
+			flags = PGrnWALApplyValueGetPositiveInteger(data, context, kv);
 		}
-		else if (PGrnWALApplyKeyEqual(context, &(kv->key), "type"))
+		else if (PGrnWALApplyKeyEqual(data, context, &(kv->key), "type"))
 		{
-			type = PGrnWALApplyValueGetGroongaObject(context, kv);
+			type = PGrnWALApplyValueGetGroongaObject(data, context, kv);
 		}
 	}
 
@@ -1740,13 +1817,13 @@ PGrnWALApplySetSources(PGrnWALApplyData *data,
 		msgpack_object_kv *kv;
 
 		kv = &(map->ptr[i]);
-		if (PGrnWALApplyKeyEqual(context, &(kv->key), "column"))
+		if (PGrnWALApplyKeyEqual(data, context, &(kv->key), "column"))
 		{
-			column = PGrnWALApplyValueGetGroongaObject(context, kv);
+			column = PGrnWALApplyValueGetGroongaObject(data, context, kv);
 		}
-		else if (PGrnWALApplyKeyEqual(context, &(kv->key), "sources"))
+		else if (PGrnWALApplyKeyEqual(data, context, &(kv->key), "sources"))
 		{
-			PGrnWALApplyValueGetGroongaObjectIDs(context, kv, sourceIDs);
+			PGrnWALApplyValueGetGroongaObjectIDs(data, context, kv, sourceIDs);
 		}
 	}
 
@@ -1769,13 +1846,13 @@ PGrnWALApplyRenameTable(PGrnWALApplyData *data,
 		msgpack_object_kv *kv;
 
 		kv = &(map->ptr[i]);
-		if (PGrnWALApplyKeyEqual(context, &(kv->key), "name"))
+		if (PGrnWALApplyKeyEqual(data, context, &(kv->key), "name"))
 		{
-			table = PGrnWALApplyValueGetGroongaObject(context, kv);
+			table = PGrnWALApplyValueGetGroongaObject(data, context, kv);
 		}
-		else if (PGrnWALApplyKeyEqual(context, &(kv->key), "new_name"))
+		else if (PGrnWALApplyKeyEqual(data, context, &(kv->key), "new_name"))
 		{
-			PGrnWALApplyValueGetString(context, kv, &newName, &newNameSize);
+			PGrnWALApplyValueGetString(data, context, kv, &newName, &newNameSize);
 		}
 	}
 
@@ -1798,13 +1875,13 @@ PGrnWALApplyDelete(PGrnWALApplyData *data,
 		msgpack_object_kv *kv;
 
 		kv = &(map->ptr[i]);
-		if (PGrnWALApplyKeyEqual(context, &(kv->key), "table"))
+		if (PGrnWALApplyKeyEqual(data, context, &(kv->key), "table"))
 		{
-			table = PGrnWALApplyValueGetGroongaObject(context, kv);
+			table = PGrnWALApplyValueGetGroongaObject(data, context, kv);
 		}
-		else if (PGrnWALApplyKeyEqual(context, &(kv->key), "key"))
+		else if (PGrnWALApplyKeyEqual(data, context, &(kv->key), "key"))
 		{
-			PGrnWALApplyValueGetBinary(context, kv, &key, &keySize);
+			PGrnWALApplyValueGetBinary(data, context, kv, &key, &keySize);
 			currentElement++;
 		}
 	}
@@ -1856,25 +1933,32 @@ PGrnWALApplyObject(PGrnWALApplyData *data, msgpack_object *object)
 		{
 		case MSGPACK_OBJECT_NIL:
 			PGrnCheckRC(GRN_INVALID_ARGUMENT,
-						"%s %s: <%u><%u>: <nil>",
+						"%s[%s(%u)] %s: <%u><%u>: <nil>",
 						tag,
+						RelationGetRelationName(data->index),
+						RelationGetRelid(data->index),
 						message,
 						currentBlock,
 						currentOffset);
 			break;
 		case MSGPACK_OBJECT_BOOLEAN:
 			PGrnCheckRC(GRN_INVALID_ARGUMENT,
-						"%s %s: <%u><%u>: <boolean>: <%s>",
+						"%s[%s(%u)] %s: <%u><%u>: <boolean>: <%s>",
 						tag,
+						RelationGetRelationName(data->index),
+						RelationGetRelid(data->index),
 						message,
 						currentBlock,
 						currentOffset,
 						object->via.boolean ? "true" : "false");
-	break;
+			break;
 		case MSGPACK_OBJECT_POSITIVE_INTEGER:
 			PGrnCheckRC(GRN_INVALID_ARGUMENT,
-						"%s %s: <%u><%u>: <positive-integer>: <%" PRId64 ">",
+						"%s[%s(%u)] %s: <%u><%u>: "
+						"<positive-integer>: <%" PRId64 ">",
 						tag,
+						RelationGetRelationName(data->index),
+						RelationGetRelid(data->index),
 						message,
 						currentBlock,
 						currentOffset,
@@ -1882,8 +1966,11 @@ PGrnWALApplyObject(PGrnWALApplyData *data, msgpack_object *object)
 			break;
 		case MSGPACK_OBJECT_NEGATIVE_INTEGER:
 			PGrnCheckRC(GRN_INVALID_ARGUMENT,
-						"%s %s: <%u><%u>: <negative-integer>: <%" PRIu64 ">",
+						"%s[%s(%u)] %s: <%u><%u>: "
+						"<negative-integer>: <%" PRIu64 ">",
 						tag,
+						RelationGetRelationName(data->index),
+						RelationGetRelid(data->index),
 						message,
 						currentBlock,
 						currentOffset,
@@ -1891,8 +1978,10 @@ PGrnWALApplyObject(PGrnWALApplyData *data, msgpack_object *object)
 			break;
 		case MSGPACK_OBJECT_FLOAT:
 			PGrnCheckRC(GRN_INVALID_ARGUMENT,
-						"%s %s: <%u><%u>: <float>: <%g>",
+						"%s[%s(%u)] %s: <%u><%u>: <float>: <%g>",
 						tag,
+						RelationGetRelationName(data->index),
+						RelationGetRelid(data->index),
 						message,
 						currentBlock,
 						currentOffset,
@@ -1900,8 +1989,10 @@ PGrnWALApplyObject(PGrnWALApplyData *data, msgpack_object *object)
 			break;
 		case MSGPACK_OBJECT_STR:
 			PGrnCheckRC(GRN_INVALID_ARGUMENT,
-						"%s %s: <%u><%u>: <string>: <%.*s>",
+						"%s[%s(%u)] %s: <%u><%u>: <string>: <%.*s>",
 						tag,
+						RelationGetRelationName(data->index),
+						RelationGetRelid(data->index),
 						message,
 						currentBlock,
 						currentOffset,
@@ -1910,8 +2001,10 @@ PGrnWALApplyObject(PGrnWALApplyData *data, msgpack_object *object)
 			break;
 		case MSGPACK_OBJECT_ARRAY:
 			PGrnCheckRC(GRN_INVALID_ARGUMENT,
-						"%s %s: <%u><%u>: <array>: <%u>",
+						"%s[%s(%u)] %s: <%u><%u>: <array>: <%u>",
 						tag,
+						RelationGetRelationName(data->index),
+						RelationGetRelid(data->index),
 						message,
 						currentBlock,
 						currentOffset,
@@ -1920,8 +2013,10 @@ PGrnWALApplyObject(PGrnWALApplyData *data, msgpack_object *object)
 #if MSGPACK_VERSION_MAJOR != 0
 		case MSGPACK_OBJECT_BIN:
 			PGrnCheckRC(GRN_INVALID_ARGUMENT,
-						"%s %s: <%u><%u>: <binary>: <%u>",
+						"%s[%s(%u)] %s: <%u><%u>: <binary>: <%u>",
 						tag,
+						RelationGetRelationName(data->index),
+						RelationGetRelid(data->index),
 						message,
 						currentBlock,
 						currentOffset,
@@ -1930,8 +2025,10 @@ PGrnWALApplyObject(PGrnWALApplyData *data, msgpack_object *object)
 #endif
 		default:
 			PGrnCheckRC(GRN_INVALID_ARGUMENT,
-						"%s %s: <%u><%u>: <%#x>",
+						"%s[%s(%u)] %s: <%u><%u>: <%#x>",
 						tag,
+						RelationGetRelationName(data->index),
+						RelationGetRelid(data->index),
 						message,
 						currentBlock,
 						currentOffset,
@@ -1947,9 +2044,9 @@ PGrnWALApplyObject(PGrnWALApplyData *data, msgpack_object *object)
 		msgpack_object_kv *kv;
 
 		kv = &(object->via.map.ptr[currentElement]);
-		if (PGrnWALApplyKeyEqual(context, &(kv->key), "_action"))
+		if (PGrnWALApplyKeyEqual(data, context, &(kv->key), "_action"))
 		{
-			action = PGrnWALApplyValueGetPositiveInteger(context, kv);
+			action = PGrnWALApplyValueGetPositiveInteger(data, context, kv);
 			currentElement++;
 		}
 	}
@@ -1976,8 +2073,10 @@ PGrnWALApplyObject(PGrnWALApplyData *data, msgpack_object *object)
 		break;
 	default:
 		PGrnCheckRC(GRN_INVALID_ARGUMENT,
-					"%s unexpected action: <%d>",
+					"%s[%s(%u)] unexpected action: <%d>",
 					tag,
+					RelationGetRelationName(data->index),
+					RelationGetRelid(data->index),
 					action);
 		break;
 	}
@@ -2033,11 +2132,15 @@ PGrnWALApplyConsume(PGrnWALApplyData *data)
 		page = BufferGetPage(buffer);
 		lastOffset = PGrnWALPageGetLastOffset(page);
 		if (dataOffset > lastOffset)
+		{
 			PGrnCheckRC(GRN_UNKNOWN_ERROR,
-						"[wal][apply][consume] "
+						"[wal][apply][consume][%s(%u)] "
 						"unconsumed WAL are overwritten: "
 						"pgroonga.max_wal_size should be increased or "
-						"pgroonga_wal_applier.naptime should be decreased");
+						"pgroonga_wal_applier.naptime should be decreased",
+						RelationGetRelationName(data->index),
+						RelationGetRelid(data->index));
+		}
 		dataSize = lastOffset - dataOffset;
 		msgpack_unpacker_reserve_buffer(&unpacker, dataSize);
 		memcpy(msgpack_unpacker_buffer(&unpacker),
@@ -2086,13 +2189,13 @@ PGrnWALApply(Relation index)
 	if (!PGrnWALApplyNeeded(&data))
 		return 0;
 
-	LockRelation(index, PGrnWALLockMode());
+	PGrnWALLock(index);
 	PGrnIndexStatusGetWALAppliedPosition(data.index,
 										 &(data.current.block),
 										 &(data.current.offset));
 	data.sources = NULL;
 	nAppliedOperations = PGrnWALApplyConsume(&data);
-	UnlockRelation(index, PGrnWALLockMode());
+	PGrnWALUnlock(index);
 #endif
 	return nAppliedOperations;
 }
@@ -2228,6 +2331,152 @@ pgroonga_wal_apply_all(PG_FUNCTION_ARGS)
 	PG_RETURN_INT64(nAppliedOperations);
 }
 
+typedef struct {
+	Relation indexes;
+	PGrnTableScanDesc scan;
+	TupleDesc desc;
+} PGrnWALStatusData;
+
+/**
+ * pgroonga_wal_status() : SETOF RECORD
+ */
+Datum
+pgroonga_wal_status(PG_FUNCTION_ARGS)
+{
+	FuncCallContext *context;
+#ifdef PGRN_SUPPORT_WAL
+	LOCKMODE lock = AccessShareLock;
+	PGrnWALStatusData *data;
+	HeapTuple indexTuple;
+
+	if (SRF_IS_FIRSTCALL())
+	{
+		MemoryContext oldContext;
+		context = SRF_FIRSTCALL_INIT();
+		oldContext = MemoryContextSwitchTo(context->multi_call_memory_ctx);
+		PG_TRY();
+		{
+			const int nAttributes = 8;
+			data = palloc(sizeof(PGrnWALStatusData));
+			data->indexes = pgrn_table_open(IndexRelationId, lock);
+			data->scan = pgrn_table_beginscan_catalog(data->indexes, 0, NULL);
+			data->desc = PGrnCreateTemplateTupleDesc(nAttributes);
+			TupleDescInitEntry(data->desc, 1, "name", TEXTOID, -1, 0);
+			TupleDescInitEntry(data->desc, 2, "oid", OIDOID, -1, 0);
+			TupleDescInitEntry(data->desc, 3, "current_block", INT8OID, -1, 0);
+			TupleDescInitEntry(data->desc, 4, "current_offset", INT8OID, -1, 0);
+			TupleDescInitEntry(data->desc, 5, "current_size", INT8OID, -1, 0);
+			TupleDescInitEntry(data->desc, 6, "last_block", INT8OID, -1, 0);
+			TupleDescInitEntry(data->desc, 7, "last_offset", INT8OID, -1, 0);
+			TupleDescInitEntry(data->desc, 8, "last_size", INT8OID, -1, 0);
+			BlessTupleDesc(data->desc);
+			context->user_fctx = data;
+		}
+		PG_CATCH();
+		{
+			MemoryContextSwitchTo(oldContext);
+			PG_RE_THROW();
+		}
+		PG_END_TRY();
+		MemoryContextSwitchTo(oldContext);
+		context->tuple_desc = data->desc;
+	}
+
+	context = SRF_PERCALL_SETUP();
+	data = context->user_fctx;
+
+	while ((indexTuple = heap_getnext(data->scan, ForwardScanDirection)))
+	{
+		Form_pg_index indexForm = (Form_pg_index) GETSTRUCT(indexTuple);
+		Relation index;
+		BlockNumber currentBlock;
+		LocationIndex currentOffset;
+		Datum *values;
+		bool *nulls;
+		int i = 0;
+
+		if (!pgrn_pg_class_ownercheck(indexForm->indexrelid, GetUserId()))
+			continue;
+
+		index = RelationIdGetRelation(indexForm->indexrelid);
+		if (!PGrnIndexIsPGroonga(index))
+		{
+			RelationClose(index);
+			continue;
+		}
+		if (PGRN_RELKIND_HAS_PARTITIONS(index->rd_rel->relkind))
+		{
+			RelationClose(index);
+			continue;
+		}
+
+		values = palloc(sizeof(Datum) * data->desc->natts);
+		/* All values are not NULL */
+		nulls = palloc0(sizeof(bool) * data->desc->natts);
+		values[i++] = PointerGetDatum(cstring_to_text(RelationGetRelationName(index)));
+		values[i++] = ObjectIdGetDatum(RelationGetRelid(index));
+		PGrnIndexStatusGetWALAppliedPosition(index, &currentBlock, &currentOffset);
+		values[i++] = Int64GetDatum(currentBlock);
+		values[i++] = Int64GetDatum(currentOffset);
+		values[i++] = Int64GetDatum(currentBlock * BLCKSZ + currentOffset);
+		if (PGrnWALEnabled && RelationGetNumberOfBlocks(index) > 0)
+		{
+			Buffer metaBuffer =
+				PGrnWALReadLockedBuffer(index,
+										PGRN_WAL_META_PAGE_BLOCK_NUMBER,
+										BUFFER_LOCK_SHARE);
+			Page metaPage = BufferGetPage(metaBuffer);
+			PGrnWALMetaPageSpecial *metaPageSpecial =
+				(PGrnWALMetaPageSpecial *) PageGetSpecialPointer(metaPage);
+			BlockNumber lastBlock = metaPageSpecial->next;
+			LocationIndex lastOffset = 0;
+			if (lastBlock < RelationGetNumberOfBlocks(index))
+			{
+				Buffer lastBuffer =
+					PGrnWALReadLockedBuffer(index,
+											metaPageSpecial->next,
+											BUFFER_LOCK_SHARE);
+				Page lastPage = BufferGetPage(lastBuffer);
+				lastOffset = PGrnWALPageGetLastOffset(lastPage);
+				UnlockReleaseBuffer(lastBuffer);
+			}
+			values[i++] = Int64GetDatum(lastBlock);
+			values[i++] = Int64GetDatum(lastOffset);
+			values[i++] = Int64GetDatum(lastBlock * BLCKSZ + lastOffset);
+			UnlockReleaseBuffer(metaBuffer);
+		}
+		else
+		{
+			values[i++] = Int64GetDatum(0);
+			values[i++] = Int64GetDatum(0);
+			values[i++] = Int64GetDatum(0);
+		}
+		RelationClose(index);
+
+		{
+			HeapTuple tuple = heap_form_tuple(data->desc, values, nulls);
+			SRF_RETURN_NEXT(context, HeapTupleGetDatum(tuple));
+		}
+	}
+
+	heap_endscan(data->scan);
+	pgrn_table_close(data->indexes, lock);
+#else
+	if (SRF_IS_FIRSTCALL())
+	{
+		context = SRF_FIRSTCALL_INIT();
+	}
+	context = SRF_PERCALL_SETUP();
+	{
+		const char *tag = "[wal][status]";
+		PGrnCheckRC(GRN_FUNCTION_NOT_IMPLEMENTED,
+					"%s not supported",
+					tag);
+	}
+#endif
+	SRF_RETURN_DONE(context);
+}
+
 #ifdef PGRN_SUPPORT_WAL
 static int64_t
 PGrnWALTruncate(Relation index)
@@ -2239,11 +2488,11 @@ PGrnWALTruncate(Relation index)
 	uint32_t nProcessingBuffers = 0;
 	GenericXLogState *state;
 
-	LockRelation(index, PGrnWALLockMode());
+	PGrnWALLock(index);
 	nBlocks = RelationGetNumberOfBlocks(index);
 	if (nBlocks == 0)
 	{
-		UnlockRelation(index, PGrnWALLockMode());
+		PGrnWALUnlock(index);
 		return 0;
 	}
 
@@ -2304,7 +2553,7 @@ PGrnWALTruncate(Relation index)
 										 PGRN_WAL_META_PAGE_BLOCK_NUMBER + 1,
 										 0);
 
-	UnlockRelation(index, PGrnWALLockMode());
+	PGrnWALUnlock(index);
 
 	return nTruncatedBlocks;
 }
